@@ -5,10 +5,13 @@ import importlib.util
 import os
 import sys
 
-from cffi import FFI
+from cffi import FFI, VerificationError
 from pywayland.ffi_build import ffi_builder as pywayland_ffi
 from xkbcommon.ffi_build import ffibuilder as xkb_ffi
 
+include_dir = (Path(__file__).parent / "include").resolve()
+assert include_dir.is_dir(), f"missing {include_dir}"
+
 
 def load_version():
     """Load the current pywlroots version"""
@@ -16,6 +19,8 @@ def load_version():
     spec = importlib.util.spec_from_file_location(
         "wlroots.version", os.path.join(dirname, "version.py")
     )
+    assert spec is not None
+    assert spec.loader is not None
     version_module = importlib.util.module_from_spec(spec)
     spec.loader.exec_module(version_module)
     return version_module.version
@@ -28,15 +33,23 @@ def load_wlroots_version():
     """
     ffi = FFI()
     ffi.cdef(CDEF_VERSION)
-    lib = ffi.verify("#include <wlr/version.h>")
+
+    try:
+        lib = ffi.verify("#include <wlr/version.h>")
+    except (PermissionError, OSError):
+        lib = importlib.import_module("wlroots").lib
 
     return f"{lib.WLR_VERSION_MAJOR}.{lib.WLR_VERSION_MINOR}.{lib.WLR_VERSION_MICRO}"
 
 
 def check_version():
     """Check for wlroots version compatibility"""
+    # When importing a system-level installed package, we may not be able to
+    # create neighboring files, which is done by the `.verify` step.  If this
+    # error is hit, we check with the compiled `wlroots.lib`.
     version = load_version()
     wlroots_version = load_wlroots_version()
+
     if version.split(".")[:2] != wlroots_version.split(".")[:2]:
         major, minor = list(map(int, version.split(".")[:2]))
         print(
@@ -45,6 +58,36 @@ def check_version():
         )
 
 
+def has_xwayland() -> bool:
+    """
+    Check for XWayland headers. If present, wlroots was built with XWayland support, so
+    pywlroots can be too.
+    """
+    try:
+        return importlib.import_module("wlroots._build").has_xwayland
+    except ModuleNotFoundError:
+        pass
+
+    try:
+        FFI().verify(
+            "#include <wlr/xwayland.h>",
+            define_macros=[("WLR_USE_UNSTABLE", 1)],
+            include_dirs=["/usr/include/pixman-1", include_dir.as_posix()],
+        )
+        has_xwayland = True
+    except VerificationError:
+        print("If XWayland support is not required, ignore the above error message.")
+        print(
+            "If support is required, ensure wlroots was built with -Dxwayland=enabled."
+        )
+        has_xwayland = False
+
+    with (Path(__file__).parent / "_build.py").open("w") as f:
+        f.write(f"has_xwayland = {has_xwayland}\n")
+
+    return has_xwayland
+
+
 # backend.h
 CDEF = """
 struct wlr_backend_impl;
@@ -65,7 +108,6 @@ struct wlr_backend *wlr_backend_autocreate(struct wl_display *display);
 
 bool wlr_backend_start(struct wlr_backend *backend);
 void wlr_backend_destroy(struct wlr_backend *backend);
-struct wlr_renderer *wlr_backend_get_renderer(struct wlr_backend *backend);
 struct wlr_session *wlr_backend_get_session(struct wlr_backend *backend);
 """
 
@@ -80,8 +122,20 @@ CDEF += """
 bool wlr_session_change_vt(struct wlr_session *session, unsigned vt);
 """
 
+# render/allocator.h
+CDEF += """
+struct wlr_allocator {
+    ...;
+};
+
+struct wlr_allocator *wlr_allocator_autocreate(struct wlr_backend *backend,
+    struct wlr_renderer *renderer);
+"""
+
 # render/wlr_renderer.h
 CDEF += """
+struct wlr_renderer *wlr_renderer_autocreate(struct wlr_backend *backend);
+
 void wlr_renderer_begin(struct wlr_renderer *r, int width, int height);
 void wlr_renderer_end(struct wlr_renderer *r);
 void wlr_renderer_clear(struct wlr_renderer *r, const float color[static 4]);
@@ -143,8 +197,6 @@ bool wlr_box_empty(const struct wlr_box *box);
 
 void wlr_box_transform(struct wlr_box *dest, const struct wlr_box *box,
     enum wl_output_transform transform, int width, int height);
-
-void wlr_box_rotated_bounds(struct wlr_box *dest, const struct wlr_box *box, float rotation);
 """
 
 # types/wlr_cursor.h
@@ -165,11 +217,14 @@ struct wlr_cursor {
         struct wl_signal pinch_begin;
         struct wl_signal pinch_update;
         struct wl_signal pinch_end;
+        struct wl_signal hold_begin;
+        struct wl_signal hold_end;
 
         struct wl_signal touch_up;
         struct wl_signal touch_down;
         struct wl_signal touch_motion;
         struct wl_signal touch_cancel;
+        struct wl_signal touch_frame;
 
         struct wl_signal tablet_tool_axis;
         struct wl_signal tablet_tool_proximity;
@@ -256,6 +311,171 @@ struct wlr_data_device_manager *wlr_data_device_manager_create(
 
 void wlr_seat_set_selection(struct wlr_seat *seat,
     struct wlr_data_source *source, uint32_t serial);
+
+void wlr_seat_start_pointer_drag(struct wlr_seat *seat, struct wlr_drag *drag,
+    uint32_t serial);
+
+struct wlr_drag_icon {
+    struct wlr_drag *drag;
+    struct wlr_surface *surface;
+    bool mapped;
+    struct {
+        struct wl_signal map;
+        struct wl_signal unmap;
+        struct wl_signal destroy;
+    } events;
+    struct wl_listener surface_destroy;
+    void *data;
+    ...;
+};
+
+struct wlr_drag {
+    enum wlr_drag_grab_type grab_type;
+    struct wlr_seat_keyboard_grab keyboard_grab;
+    struct wlr_seat_pointer_grab pointer_grab;
+    struct wlr_seat_touch_grab touch_grab;
+    struct wlr_seat *seat;
+    struct wlr_seat_client *seat_client;
+    struct wlr_seat_client *focus_client;
+    struct wlr_drag_icon *icon; // can be NULL
+    struct wlr_surface *focus; // can be NULL
+    struct wlr_data_source *source; // can be NULL
+    bool started, dropped, cancelling;
+    int32_t grab_touch_id, touch_id; // if WLR_DRAG_GRAB_TOUCH
+    struct {
+        struct wl_signal focus;
+        struct wl_signal motion; // wlr_drag_motion_event
+        struct wl_signal drop; // wlr_drag_drop_event
+        struct wl_signal destroy;
+    } events;
+    struct wl_listener source_destroy;
+    struct wl_listener seat_client_destroy;
+    struct wl_listener icon_destroy;
+    void *data;
+    ...;
+};
+
+struct wlr_drag_motion_event {
+    struct wlr_drag *drag;
+    uint32_t time;
+    double sx, sy;
+    ...;
+};
+
+struct wlr_drag_drop_event {
+    struct wlr_drag *drag;
+    uint32_t time;
+    ...;
+};
+
+struct wlr_data_source {
+    const struct wlr_data_source_impl *impl;
+    struct wl_array mime_types;
+    int32_t actions;
+    bool accepted;
+    enum wl_data_device_manager_dnd_action current_dnd_action;
+    uint32_t compositor_action;
+    struct {
+        struct wl_signal destroy;
+    } events;
+    ...;
+};
+
+void wlr_data_source_destroy(struct wlr_data_source *source);
+"""
+
+# types/wlr_foreign_toplevel_management_v1.h
+CDEF += """
+struct wlr_foreign_toplevel_manager_v1 {
+    struct wl_event_loop *event_loop;
+    struct wl_global *global;
+    struct wl_list resources; // wl_resource_get_link
+    struct wl_list toplevels; // wlr_foreign_toplevel_handle_v1::link
+
+    struct wl_listener display_destroy;
+
+    struct {
+        struct wl_signal destroy;
+    } events;
+
+    void *data;
+    ...;
+};
+
+struct wlr_foreign_toplevel_handle_v1 {
+    struct wlr_foreign_toplevel_manager_v1 *manager;
+    struct wl_list resources;
+    struct wl_list link;
+    struct wl_event_source *idle_source;
+
+    char *title;
+    char *app_id;
+    struct wlr_foreign_toplevel_handle_v1 *parent;
+    struct wl_list outputs; // wlr_foreign_toplevel_v1_output
+    uint32_t state; // wlr_foreign_toplevel_v1_state
+
+    struct {
+        struct wl_signal request_maximize;
+        struct wl_signal request_minimize;
+        struct wl_signal request_activate;
+        struct wl_signal request_fullscreen;
+        struct wl_signal request_close;
+        struct wl_signal set_rectangle;
+        struct wl_signal destroy;
+    } events;
+
+    void *data;
+    ...;
+};
+
+struct wlr_foreign_toplevel_handle_v1_maximized_event {
+    struct wlr_foreign_toplevel_handle_v1 *toplevel;
+    bool maximized;
+};
+struct wlr_foreign_toplevel_handle_v1_minimized_event {
+    struct wlr_foreign_toplevel_handle_v1 *toplevel;
+    bool minimized;
+};
+struct wlr_foreign_toplevel_handle_v1_activated_event {
+    struct wlr_foreign_toplevel_handle_v1 *toplevel;
+    struct wlr_seat *seat;
+};
+struct wlr_foreign_toplevel_handle_v1_fullscreen_event {
+    struct wlr_foreign_toplevel_handle_v1 *toplevel;
+    bool fullscreen;
+    struct wlr_output *output;
+};
+struct wlr_foreign_toplevel_handle_v1_set_rectangle_event {
+    struct wlr_foreign_toplevel_handle_v1 *toplevel;
+    struct wlr_surface *surface;
+    int32_t x, y, width, height;
+};
+
+struct wlr_foreign_toplevel_manager_v1 *wlr_foreign_toplevel_manager_v1_create(
+    struct wl_display *display);
+struct wlr_foreign_toplevel_handle_v1 *wlr_foreign_toplevel_handle_v1_create(
+    struct wlr_foreign_toplevel_manager_v1 *manager);
+void wlr_foreign_toplevel_handle_v1_destroy(
+    struct wlr_foreign_toplevel_handle_v1 *toplevel);
+void wlr_foreign_toplevel_handle_v1_set_title(
+    struct wlr_foreign_toplevel_handle_v1 *toplevel, const char *title);
+void wlr_foreign_toplevel_handle_v1_set_app_id(
+    struct wlr_foreign_toplevel_handle_v1 *toplevel, const char *app_id);
+void wlr_foreign_toplevel_handle_v1_output_enter(
+    struct wlr_foreign_toplevel_handle_v1 *toplevel, struct wlr_output *output);
+void wlr_foreign_toplevel_handle_v1_output_leave(
+    struct wlr_foreign_toplevel_handle_v1 *toplevel, struct wlr_output *output);
+void wlr_foreign_toplevel_handle_v1_set_maximized(
+    struct wlr_foreign_toplevel_handle_v1 *toplevel, bool maximized);
+void wlr_foreign_toplevel_handle_v1_set_minimized(
+    struct wlr_foreign_toplevel_handle_v1 *toplevel, bool minimized);
+void wlr_foreign_toplevel_handle_v1_set_activated(
+    struct wlr_foreign_toplevel_handle_v1 *toplevel, bool activated);
+void wlr_foreign_toplevel_handle_v1_set_fullscreen(
+    struct wlr_foreign_toplevel_handle_v1* toplevel, bool fullscreen);
+void wlr_foreign_toplevel_handle_v1_set_parent(
+    struct wlr_foreign_toplevel_handle_v1 *toplevel,
+    struct wlr_foreign_toplevel_handle_v1 *parent);
 """
 
 # types/wlr_gamma_control_v1.h
@@ -277,6 +497,108 @@ struct wlr_gamma_control_manager_v1 *wlr_gamma_control_manager_v1_create(
     struct wl_display *display);
 """
 
+# types/wlr_idle.h
+CDEF += """
+struct wlr_idle {
+    struct wl_global *global;
+    struct wl_list idle_timers; // wlr_idle_timeout::link
+    struct wl_event_loop *event_loop;
+    bool enabled;
+
+    struct wl_listener display_destroy;
+    struct {
+        struct wl_signal activity_notify;
+        struct wl_signal destroy;
+    } events;
+
+    void *data;
+    ...;
+};
+
+struct wlr_idle_timeout {
+    struct wl_resource *resource;
+    struct wl_list link;
+    struct wlr_seat *seat;
+
+    struct wl_event_source *idle_source;
+    bool idle_state;
+    bool enabled;
+    uint32_t timeout; // milliseconds
+
+    struct {
+        struct wl_signal idle;
+        struct wl_signal resume;
+        struct wl_signal destroy;
+    } events;
+
+    struct wl_listener input_listener;
+    struct wl_listener seat_destroy;
+
+    void *data;
+    ...;
+};
+
+struct wlr_idle *wlr_idle_create(struct wl_display *display);
+
+/**
+ * Send notification to restart all timers for the given seat. Called by
+ * compositor when there is an user activity event on that seat.
+ */
+void wlr_idle_notify_activity(struct wlr_idle *idle, struct wlr_seat *seat);
+
+/**
+ * Enable or disable timers for a given idle resource by seat.
+ * Passing a NULL seat means update timers for all seats.
+ */
+void wlr_idle_set_enabled(struct wlr_idle *idle, struct wlr_seat *seat,
+    bool enabled);
+
+/**
+ * Create a new timer on the given seat. The idle event will be called after
+ * the given amount of milliseconds of inactivity, and the resumed event will
+ * be sent at the first user activity after the fired event.
+ */
+struct wlr_idle_timeout *wlr_idle_timeout_create(struct wlr_idle *idle,
+    struct wlr_seat *seat, uint32_t timeout);
+
+void wlr_idle_timeout_destroy(struct wlr_idle_timeout *timeout);
+"""
+
+# types/wlr_input_inhibit_v1.h
+CDEF += """
+struct wlr_idle_inhibit_manager_v1 {
+    struct wl_list inhibitors; // wlr_idle_inhibit_inhibitor_v1::link
+    struct wl_global *global;
+
+    struct wl_listener display_destroy;
+
+    struct {
+        struct wl_signal new_inhibitor;
+        struct wl_signal destroy;
+    } events;
+
+    void *data;
+    ...;
+};
+
+struct wlr_idle_inhibitor_v1 {
+    struct wlr_surface *surface;
+    struct wl_resource *resource;
+    struct wl_listener surface_destroy;
+
+    struct wl_list link; // wlr_idle_inhibit_manager_v1::inhibitors;
+
+    struct {
+        struct wl_signal destroy;
+    } events;
+
+    void *data;
+    ...;
+};
+
+struct wlr_idle_inhibit_manager_v1 *wlr_idle_inhibit_v1_create(struct wl_display *display);
+"""
+
 # types/wlr_input_device.h
 CDEF += """
 enum wlr_button_state {
@@ -320,8 +642,6 @@ struct wlr_input_device {
     } events;
 
     void *data;
-
-    struct wl_list link;
     ...;
 };
 """
@@ -497,7 +817,8 @@ struct wlr_output {
     struct wl_global *global;
     struct wl_list resources;
 
-    char name[24];
+    char *name;
+    char *description; // may be NULL
     char make[56];
     char model[16];
     char serial[16];
@@ -553,6 +874,9 @@ void wlr_output_enable(struct wlr_output *output, bool enable);
 void wlr_output_create_global(struct wlr_output *output);
 void wlr_output_destroy_global(struct wlr_output *output);
 
+bool wlr_output_init_render(struct wlr_output *output,
+    struct wlr_allocator *allocator, struct wlr_renderer *renderer);
+
 struct wlr_output_mode *wlr_output_preferred_mode(struct wlr_output *output);
 
 void wlr_output_set_mode(struct wlr_output *output,
@@ -595,7 +919,7 @@ struct wlr_output_damage {
     struct pixman_region32 previous[WLR_OUTPUT_DAMAGE_PREVIOUS_LEN];
     size_t previous_idx;
 
-    enum wlr_output_state_buffer_type pending_buffer_type;
+    bool pending_attach_render;
 
     struct {
         struct wl_signal frame;
@@ -746,6 +1070,46 @@ void wlr_output_configuration_v1_destroy(
     struct wlr_output_configuration_v1 *config);
 """
 
+# types/wlr_output_powewr_management_v1.h
+CDEF += """
+struct wlr_output_power_manager_v1 {
+    struct wl_global *global;
+    struct wl_list output_powers; // wlr_output_power_v1::link
+
+    struct wl_listener display_destroy;
+
+    struct {
+        struct wl_signal set_mode; // wlr_output_power_v1_set_mode_event
+        struct wl_signal destroy;
+    } events;
+
+    void *data;
+    ...;
+};
+
+struct wlr_output_power_v1 {
+    struct wl_resource *resource;
+    struct wlr_output *output;
+    struct wlr_output_power_manager_v1 *manager;
+    struct wl_list link;
+
+    struct wl_listener output_destroy_listener;
+    struct wl_listener output_commit_listener;
+
+    void *data;
+    ...;
+};
+
+struct wlr_output_power_v1_set_mode_event {
+    struct wlr_output *output;
+    enum zwlr_output_power_v1_mode mode;
+    ...;
+};
+
+struct wlr_output_power_manager_v1 *wlr_output_power_manager_v1_create(
+    struct wl_display *display);
+"""
+
 # types/wlr_pointer.h
 CDEF += """
 struct wlr_pointer {
@@ -869,11 +1233,13 @@ CDEF += """
 enum wlr_pointer_constraint_v1_type {
     WLR_POINTER_CONSTRAINT_V1_LOCKED,
     WLR_POINTER_CONSTRAINT_V1_CONFINED,
+    ...
 };
 
 enum wlr_pointer_constraint_v1_state_field {
     WLR_POINTER_CONSTRAINT_V1_STATE_REGION = 1 << 0,
     WLR_POINTER_CONSTRAINT_V1_STATE_CURSOR_HINT = 1 << 1,
+    ...
 };
 
 struct wlr_pointer_constraint_v1_state {
@@ -951,6 +1317,7 @@ struct wlr_primary_selection_v1_device_manager {
     } events;
 
     void *data;
+    ...;
 };
 struct wlr_primary_selection_v1_device_manager *
     wlr_primary_selection_v1_device_manager_create(struct wl_display *display);
@@ -990,6 +1357,57 @@ void wlr_relative_pointer_manager_v1_send_relative_motion(
     double dx_unaccel, double dy_unaccel);
 """
 
+# types/wlr_scene.h
+CDEF += """
+struct wlr_scene_node_state {
+    struct wl_list link; // wlr_scene_node_state.children
+    struct wl_list children; // wlr_scene_node_state.link
+
+    bool enabled;
+    int x, y; // relative to parent
+};
+
+struct wlr_scene_node {
+    enum wlr_scene_node_type type;
+    struct wlr_scene_node *parent;
+    struct wlr_scene_node_state state;
+
+    struct {
+        struct wl_signal destroy;
+    } events;
+
+    void *data;
+    ...;
+};
+
+struct wlr_scene {
+    struct wlr_scene_node node;
+
+    struct wl_list outputs; // wlr_scene_output.link
+
+    ...;
+};
+
+void wlr_scene_node_set_position(struct wlr_scene_node *node, int x, int y);
+
+void wlr_scene_node_place_above(struct wlr_scene_node *node, struct wlr_scene_node *sibling);
+void wlr_scene_node_place_below(struct wlr_scene_node *node, struct wlr_scene_node *sibling);
+void wlr_scene_node_raise_to_top(struct wlr_scene_node *node);
+void wlr_scene_node_lower_to_bottom(struct wlr_scene_node *node);
+
+struct wlr_scene_node *wlr_scene_xdg_surface_create(
+    struct wlr_scene_node *parent, struct wlr_xdg_surface *xdg_surface);
+
+bool wlr_scene_output_commit(struct wlr_scene_output *scene_output);
+void wlr_scene_output_send_frame_done(struct wlr_scene_output *scene_output, struct timespec *now);
+struct wlr_scene_output *wlr_scene_get_scene_output(struct wlr_scene *scene, struct wlr_output *output);
+
+struct wlr_scene *wlr_scene_create(void);
+
+bool wlr_scene_attach_output_layout(struct wlr_scene *scene,
+    struct wlr_output_layout *output_layout);
+"""
+
 # types/wlr_screencopy_v1.h
 CDEF += """
 struct wlr_screencopy_manager_v1 {
@@ -1024,6 +1442,25 @@ int clock_gettime(clockid_t clk_id, struct timespec *tp);
 
 #define WLR_POINTER_BUTTONS_CAP 16
 
+struct wlr_seat_touch_grab {
+    const struct wlr_touch_grab_interface *interface;
+    struct wlr_seat *seat;
+    void *data;
+    ...;
+};
+struct wlr_seat_keyboard_grab {
+    const struct wlr_keyboard_grab_interface *interface;
+    struct wlr_seat *seat;
+    void *data;
+    ...;
+};
+struct wlr_seat_pointer_grab {
+    const struct wlr_pointer_grab_interface *interface;
+    struct wlr_seat *seat;
+    void *data;
+    ...;
+};
+
 struct wlr_seat_pointer_state {
     struct wlr_seat *seat;
     struct wlr_seat_client *focused_client;
@@ -1136,28 +1573,33 @@ struct wlr_seat_pointer_request_set_cursor_event {
     struct wlr_surface *surface;
     uint32_t serial;
     int32_t hotspot_x, hotspot_y;
+    ...;
 };
 
 struct wlr_seat_request_set_selection_event {
     struct wlr_data_source *source;
     uint32_t serial;
+    ...;
 };
 
 struct wlr_seat_request_start_drag_event {
     struct wlr_drag *drag;
     struct wlr_surface *origin;
     uint32_t serial;
+    ...;
 };
 
 struct wlr_seat_pointer_focus_change_event {
     struct wlr_seat *seat;
     struct wlr_surface *old_surface, *new_surface;
     double sx, sy;
+    ...;
 };
 
 struct wlr_seat_keyboard_focus_change_event {
     struct wlr_seat *seat;
     struct wlr_surface *old_surface, *new_surface;
+    ...;
 };
 
 struct wlr_seat *wlr_seat_create(struct wl_display *display, const char *name);
@@ -1189,6 +1631,14 @@ bool wlr_seat_pointer_has_grab(struct wlr_seat *seat);
 void wlr_seat_set_keyboard(struct wlr_seat *seat, struct wlr_input_device *dev);
 struct wlr_keyboard *wlr_seat_get_keyboard(struct wlr_seat *seat);
 
+uint32_t wlr_seat_touch_notify_down(struct wlr_seat *seat,
+    struct wlr_surface *surface, uint32_t time_msec,
+    int32_t touch_id, double sx, double sy);
+void wlr_seat_touch_notify_up(struct wlr_seat *seat, uint32_t time_msec,
+    int32_t touch_id);
+void wlr_seat_touch_notify_motion(struct wlr_seat *seat, uint32_t time_msec,
+    int32_t touch_id, double sx, double sy);
+
 void wlr_seat_keyboard_start_grab(struct wlr_seat *wlr_seat,
     struct wlr_seat_keyboard_grab *grab);
 void wlr_seat_keyboard_end_grab(struct wlr_seat *wlr_seat);
@@ -1200,7 +1650,10 @@ void wlr_seat_keyboard_notify_enter(struct wlr_seat *seat,
     struct wlr_surface *surface, uint32_t keycodes[], size_t num_keycodes,
     struct wlr_keyboard_modifiers *modifiers);
 void wlr_seat_keyboard_clear_focus(struct wlr_seat *wlr_seat);
+void wlr_seat_pointer_notify_clear_focus(struct wlr_seat *wlr_seat);
 bool wlr_seat_keyboard_has_grab(struct wlr_seat *seat);
+bool wlr_seat_validate_pointer_grab_serial(struct wlr_seat *seat,
+    struct wlr_surface *origin, uint32_t serial);
 """
 
 # types/wlr_server_decoration.h
@@ -1221,6 +1674,7 @@ struct wlr_server_decoration_manager {
         struct wl_signal destroy;
     } events;
     void *data;
+    ...;
 };
 struct wlr_server_decoration_manager *wlr_server_decoration_manager_create(
     struct wl_display *display);
@@ -1234,7 +1688,7 @@ struct wlr_surface_state {
     uint32_t committed;
     uint32_t seq;
 
-    struct wl_resource *buffer_resource;
+    struct wlr_buffer *buffer;
     int32_t dx, dy;
     struct pixman_region32 surface_damage, buffer_damage;
     struct pixman_region32 opaque, input;
@@ -1245,6 +1699,8 @@ struct wlr_surface_state {
     int width, height;
     int buffer_width, buffer_height;
 
+    struct wl_list subsurfaces_below;
+    struct wl_list subsurfaces_above;
     ...;
 };
 
@@ -1272,7 +1728,7 @@ struct wlr_surface {
     struct pixman_region32 buffer_damage;
     struct pixman_region32 opaque_region;
     struct pixman_region32 input_region;
-    struct wlr_surface_state current, pending, previous;
+    struct wlr_surface_state current, pending;
 
     struct wl_list cached;
 
@@ -1285,18 +1741,14 @@ struct wlr_surface {
         struct wl_signal destroy;
     } events;
 
-    struct wl_list subsurfaces_below;
-    struct wl_list subsurfaces_above;
-    struct wl_list subsurfaces_pending_below;
-    struct wl_list subsurfaces_pending_above;
     struct wl_list current_outputs;
+    void *data;
     struct wl_listener renderer_destroy;
 
-    void *data;
     ...;
 };
 
-struct wlr_subsurface_state {
+struct wlr_subsurface_parent_state {
     int32_t x, y;
     ...;
 };
@@ -1306,7 +1758,7 @@ struct wlr_subsurface {
     struct wlr_surface *surface;
     struct wlr_surface *parent;
 
-    struct wlr_subsurface_state current, pending;
+    struct wlr_subsurface_parent_state current, pending;
 
     uint32_t cached_seq;
     bool has_cache;
@@ -1315,9 +1767,6 @@ struct wlr_subsurface {
     bool reordered;
     bool mapped;
 
-    struct wl_list parent_link;
-    struct wl_list parent_pending_link;
-
     struct wl_listener surface_destroy;
     struct wl_listener parent_destroy;
 
@@ -1342,10 +1791,6 @@ bool wlr_surface_has_buffer(struct wlr_surface *surface);
 
 struct wlr_texture *wlr_surface_get_texture(struct wlr_surface *surface);
 
-struct wlr_subsurface *wlr_subsurface_create(struct wlr_surface *surface,
-    struct wlr_surface *parent, uint32_t version, uint32_t id,
-    struct wl_list *resource_list);
-
 struct wlr_surface *wlr_surface_get_root_surface(struct wlr_surface *surface);
 
 bool wlr_surface_point_accepts_input(struct wlr_surface *surface,
@@ -1426,6 +1871,26 @@ int wlr_xcursor_manager_load(struct wlr_xcursor_manager *manager,
     float scale);
 void wlr_xcursor_manager_set_cursor_image(struct wlr_xcursor_manager *manager,
     const char *name, struct wlr_cursor *cursor);
+struct wlr_xcursor *wlr_xcursor_manager_get_xcursor(
+    struct wlr_xcursor_manager *manager, const char *name, float scale);
+"""
+
+# xcursor.h
+CDEF += """
+struct wlr_xcursor_image {
+    uint32_t width;
+    uint32_t height;
+    uint32_t hotspot_x;
+    uint32_t hotspot_y;
+    uint32_t delay;
+    uint8_t *buffer;
+};
+struct wlr_xcursor {
+    unsigned int image_count;
+    struct wlr_xcursor_image **images;
+    char *name;
+    uint32_t total_delay;
+};
 """
 
 # types/wlr_xdg_decoration_v1.h
@@ -1455,6 +1920,12 @@ struct wlr_xdg_toplevel_decoration_v1_configure {
     struct wl_list link; // wlr_xdg_toplevel_decoration::configure_list
     struct wlr_xdg_surface_configure *surface_configure;
     enum wlr_xdg_toplevel_decoration_v1_mode mode;
+    ...;
+};
+
+struct wlr_xdg_toplevel_decoration_v1_state {
+    enum wlr_xdg_toplevel_decoration_v1_mode mode;
+    ...;
 };
 
 struct wlr_xdg_toplevel_decoration_v1 {
@@ -1463,10 +1934,12 @@ struct wlr_xdg_toplevel_decoration_v1 {
     struct wlr_xdg_decoration_manager_v1 *manager;
     struct wl_list link; // wlr_xdg_decoration_manager_v1::link
 
+    struct wlr_xdg_toplevel_decoration_v1_state current, pending;
+
+    enum wlr_xdg_toplevel_decoration_v1_mode scheduled_mode;
+    enum wlr_xdg_toplevel_decoration_v1_mode requested_mode;
+
     bool added;
-    enum wlr_xdg_toplevel_decoration_v1_mode current_mode;
-    enum wlr_xdg_toplevel_decoration_v1_mode client_pending_mode;
-    enum wlr_xdg_toplevel_decoration_v1_mode server_pending_mode;
 
     struct wl_list configure_list; // wlr_xdg_toplevel_decoration_v1_configure::link
 
@@ -1544,6 +2017,7 @@ struct wlr_xdg_client {
 
     uint32_t ping_serial;
     struct wl_event_source *ping_timer;
+    ...;
 };
 
 struct wlr_xdg_positioner {
@@ -1583,11 +2057,22 @@ struct wlr_xdg_shell *wlr_xdg_shell_create(struct wl_display *display);
 
 struct wlr_xdg_toplevel_state {
     bool maximized, fullscreen, resizing, activated;
-    uint32_t tiled;
+    uint32_t tiled; // enum wlr_edges
     uint32_t width, height;
     uint32_t max_width, max_height;
     uint32_t min_width, min_height;
+    ...;
+};
 
+struct wlr_xdg_toplevel_configure {
+    bool maximized, fullscreen, resizing, activated;
+    uint32_t tiled; // enum wlr_edges
+    uint32_t width, height;
+    ...;
+};
+
+struct wlr_xdg_toplevel_requested {
+    bool maximized, minimized, fullscreen;
     struct wlr_output *fullscreen_output;
     struct wl_listener fullscreen_output_destroy;
     ...;
@@ -1601,9 +2086,11 @@ struct wlr_xdg_toplevel {
     struct wlr_xdg_surface *parent;
     struct wl_listener parent_unmap;
 
-    struct wlr_xdg_toplevel_state client_pending;
-    struct wlr_xdg_toplevel_state server_pending;
-    struct wlr_xdg_toplevel_state current;
+    struct wlr_xdg_toplevel_state current, pending;
+
+    struct wlr_xdg_toplevel_configure scheduled;
+
+    struct wlr_xdg_toplevel_requested requested;
 
     char *title;
     char *app_id;
@@ -1644,14 +2131,11 @@ struct wlr_xdg_surface {
     struct wl_list popups; // wlr_xdg_popup::link
 
     bool added, configured, mapped;
-    uint32_t configure_serial;
     struct wl_event_source *configure_idle;
-    uint32_t configure_next_serial;
+    uint32_t scheduled_serial;
     struct wl_list configure_list;
 
-    bool has_next_geometry;
-    struct wlr_box next_geometry;
-    struct wlr_box geometry;
+    struct wlr_xdg_surface_state current, pending;
 
     struct wl_listener surface_destroy;
     struct wl_listener surface_commit;
@@ -1676,7 +2160,14 @@ struct wlr_xdg_surface_configure {
     struct wl_list link; // wlr_xdg_surface::configure_list
     uint32_t serial;
 
-    struct wlr_xdg_toplevel_state *toplevel_state;
+    struct wlr_xdg_toplevel_configure *toplevel_configure;
+    ...;
+};
+
+struct wlr_xdg_surface_state {
+    uint32_t configure_serial;
+    struct wlr_box geometry;
+    ...;
 };
 
 struct wlr_xdg_toplevel_move_event {
@@ -1787,9 +2278,6 @@ void wlr_region_transform(struct pixman_region32 *dst, struct pixman_region32 *s
 CDEF += """
 struct wlr_backend *wlr_headless_backend_create(struct wl_display *display);
 
-struct wlr_backend *wlr_headless_backend_create_with_renderer(
-    struct wl_display *display, struct wlr_renderer *renderer);
-
 struct wlr_output *wlr_headless_add_output(struct wlr_backend *backend,
     unsigned int width, unsigned int height);
 
@@ -1813,12 +2301,16 @@ SOURCE = """
 #include <wlr/backend.h>
 #include <wlr/backend/headless.h>
 #include <wlr/backend/libinput.h>
+#include <wlr/render/allocator.h>
 #include <wlr/render/wlr_renderer.h>
 #include <wlr/types/wlr_cursor.h>
 #include <wlr/types/wlr_compositor.h>
 #include <wlr/types/wlr_data_control_v1.h>
 #include <wlr/types/wlr_data_device.h>
+#include <wlr/types/wlr_foreign_toplevel_management_v1.h>
 #include <wlr/types/wlr_gamma_control_v1.h>
+#include <wlr/types/wlr_idle.h>
+#include <wlr/types/wlr_idle_inhibit_v1.h>
 #include <wlr/types/wlr_input_inhibitor.h>
 #include <wlr/types/wlr_keyboard.h>
 #include <wlr/types/wlr_layer_shell_v1.h>
@@ -1828,9 +2320,11 @@ SOURCE = """
 #include <wlr/types/wlr_output_damage.h>
 #include <wlr/types/wlr_output_layout.h>
 #include <wlr/types/wlr_output_management_v1.h>
+#include <wlr/types/wlr_output_power_management_v1.h>
 #include <wlr/types/wlr_pointer_constraints_v1.h>
 #include <wlr/types/wlr_primary_selection_v1.h>
 #include <wlr/types/wlr_relative_pointer_v1.h>
+#include <wlr/types/wlr_scene.h>
 #include <wlr/types/wlr_screencopy_v1.h>
 #include <wlr/types/wlr_surface.h>
 #include <wlr/types/wlr_seat.h>
@@ -1914,14 +2408,9 @@ struct wlr_layer_surface_v1 {
     struct wlr_layer_shell_v1 *shell;
     struct wl_list popups; // wlr_xdg_popup::link
     char *namespace;
-    bool added, configured, mapped, closed;
-    uint32_t configure_serial;
-    uint32_t configure_next_serial;
+    bool added, configured, mapped;
     struct wl_list configure_list;
-    struct wlr_layer_surface_v1_configure *acked_configure;
-    struct wlr_layer_surface_v1_state client_pending;
-    struct wlr_layer_surface_v1_state server_pending;
-    struct wlr_layer_surface_v1_state current;
+    struct wlr_layer_surface_v1_state current, pending;
     struct wl_listener surface_destroy;
     struct {
         struct wl_signal destroy;
@@ -1939,7 +2428,7 @@ struct wlr_layer_shell_v1 *wlr_layer_shell_v1_create(struct wl_display *display)
 void wlr_layer_surface_v1_configure(struct wlr_layer_surface_v1 *surface,
     uint32_t width, uint32_t height);
 
-void wlr_layer_surface_v1_close(struct wlr_layer_surface_v1 *surface);
+void wlr_layer_surface_v1_destroy(struct wlr_layer_surface_v1 *surface);
 
 bool wlr_surface_is_layer_surface(struct wlr_surface *surface);
 
@@ -1956,8 +2445,255 @@ struct wlr_surface *wlr_layer_surface_v1_surface_at(
 
 check_version()
 
-include_dir = (Path(__file__).parent.parent / "include").resolve()
-assert include_dir.is_dir(), f"missing {include_dir}"
+if has_xwayland():
+    # xwayland.h
+    print("Built with support for XWayland helpers.")
+
+    SOURCE += """
+    #include <wlr/xwayland.h>
+    """
+
+    CDEF += """
+    typedef int... time_t;
+    typedef struct {
+        ...;
+    } xcb_generic_event_t;
+    typedef uint32_t xcb_pixmap_t;
+    typedef uint32_t xcb_window_t;
+    typedef uint32_t xcb_atom_t;
+
+    struct wlr_xwm;
+    struct wlr_xwayland_cursor;
+    struct wlr_xwayland_server_options {
+        bool lazy;
+        bool enable_wm;
+        bool no_touch_pointer_emulation;
+        ...;
+    };
+    struct wlr_xwayland_server {
+        pid_t pid;
+        struct wl_client *client;
+        struct wl_event_source *pipe_source;
+        int wm_fd[2], wl_fd[2];
+        time_t server_start;
+        int display;
+        char display_name[16];
+        int x_fd[2];
+        struct wl_event_source *x_fd_read_event[2];
+        struct wlr_xwayland_server_options options;
+        struct wl_display *wl_display;
+        struct {
+            struct wl_signal ready;
+            struct wl_signal destroy;
+        } events;
+        struct wl_listener client_destroy;
+        struct wl_listener display_destroy;
+        void *data;
+        ...;
+    };
+    struct wlr_xwayland {
+        struct wlr_xwayland_server *server;
+        struct wlr_xwm *xwm;
+        struct wlr_xwayland_cursor *cursor;
+        const char *display_name;
+        struct wl_display *wl_display;
+        struct wlr_compositor *compositor;
+        struct wlr_seat *seat;
+        struct {
+            struct wl_signal ready;
+            struct wl_signal new_surface;
+            struct wl_signal remove_startup_info;
+        } events;
+        int (*user_event_handler)(struct wlr_xwm *xwm, xcb_generic_event_t *event);
+        struct wl_listener server_ready;
+        struct wl_listener server_destroy;
+        struct wl_listener seat_destroy;
+        void *data;
+        ...;
+    };
+    enum wlr_xwayland_surface_decorations {
+        WLR_XWAYLAND_SURFACE_DECORATIONS_ALL = 0,
+        WLR_XWAYLAND_SURFACE_DECORATIONS_NO_BORDER = 1,
+        WLR_XWAYLAND_SURFACE_DECORATIONS_NO_TITLE = 2,
+    };
+    struct wlr_xwayland_surface_hints {
+        uint32_t flags;
+        uint32_t input;
+        int32_t initial_state;
+        xcb_pixmap_t icon_pixmap;
+        xcb_window_t icon_window;
+        int32_t icon_x, icon_y;
+        xcb_pixmap_t icon_mask;
+        xcb_window_t window_group;
+        ...;
+    };
+    struct wlr_xwayland_surface_size_hints {
+        uint32_t flags;
+        int32_t x, y;
+        int32_t width, height;
+        int32_t min_width, min_height;
+        int32_t max_width, max_height;
+        int32_t width_inc, height_inc;
+        int32_t base_width, base_height;
+        int32_t min_aspect_num, min_aspect_den;
+        int32_t max_aspect_num, max_aspect_den;
+        uint32_t win_gravity;
+        ...;
+    };
+    enum wlr_xwayland_icccm_input_model {
+        WLR_ICCCM_INPUT_MODEL_NONE = 0,
+        WLR_ICCCM_INPUT_MODEL_PASSIVE = 1,
+        WLR_ICCCM_INPUT_MODEL_LOCAL = 2,
+        WLR_ICCCM_INPUT_MODEL_GLOBAL = 3,
+    };
+    struct wlr_xwayland_surface {
+        xcb_window_t window_id;
+        struct wlr_xwm *xwm;
+        uint32_t surface_id;
+        struct wl_list link;
+        struct wl_list stack_link;
+        struct wl_list unpaired_link;
+        struct wlr_surface *surface;
+        int16_t x, y;
+        uint16_t width, height;
+        uint16_t saved_width, saved_height;
+        bool override_redirect;
+        bool mapped;
+        char *title;
+        char *class;
+        char *instance;
+        char *role;
+        char *startup_id;
+        pid_t pid;
+        bool has_utf8_title;
+        struct wl_list children; // wlr_xwayland_surface::parent_link
+        struct wlr_xwayland_surface *parent;
+        struct wl_list parent_link; // wlr_xwayland_surface::children
+        xcb_atom_t *window_type;
+        size_t window_type_len;
+        xcb_atom_t *protocols;
+        size_t protocols_len;
+        uint32_t decorations;
+        struct wlr_xwayland_surface_hints *hints;
+        uint32_t hints_urgency;
+        struct wlr_xwayland_surface_size_hints *size_hints;
+        bool pinging;
+        struct wl_event_source *ping_timer;
+        // _NET_WM_STATE
+        bool modal;
+        bool fullscreen;
+        bool maximized_vert, maximized_horz;
+        bool minimized;
+        bool has_alpha;
+        struct {
+            struct wl_signal destroy;
+            struct wl_signal request_configure;
+            struct wl_signal request_move;
+            struct wl_signal request_resize;
+            struct wl_signal request_minimize;
+            struct wl_signal request_maximize;
+            struct wl_signal request_fullscreen;
+            struct wl_signal request_activate;
+            struct wl_signal map;
+            struct wl_signal unmap;
+            struct wl_signal set_title;
+            struct wl_signal set_class;
+            struct wl_signal set_role;
+            struct wl_signal set_parent;
+            struct wl_signal set_pid;
+            struct wl_signal set_startup_id;
+            struct wl_signal set_window_type;
+            struct wl_signal set_hints;
+            struct wl_signal set_decorations;
+            struct wl_signal set_override_redirect;
+            struct wl_signal set_geometry;
+            struct wl_signal ping_timeout;
+        } events;
+        struct wl_listener surface_destroy;
+        void *data;
+        ...;
+    };
+    struct wlr_xwayland_surface_configure_event {
+        struct wlr_xwayland_surface *surface;
+        int16_t x, y;
+        uint16_t width, height;
+        uint16_t mask; // xcb_config_window_t
+        ...;
+    };
+    struct wlr_xwayland_move_event {
+        struct wlr_xwayland_surface *surface;
+        ...;
+    };
+    struct wlr_xwayland_remove_startup_info_event  {
+        const char *id;
+        xcb_window_t window;
+        ...;
+    };
+    struct wlr_xwayland_resize_event {
+        struct wlr_xwayland_surface *surface;
+        uint32_t edges;
+        ...;
+    };
+    struct wlr_xwayland_minimize_event {
+        struct wlr_xwayland_surface *surface;
+        bool minimize;
+        ...;
+    };
+    struct wlr_xwayland_server *wlr_xwayland_server_create(
+        struct wl_display *display, struct wlr_xwayland_server_options *options);
+    void wlr_xwayland_server_destroy(struct wlr_xwayland_server *server);
+    struct wlr_xwayland *wlr_xwayland_create(struct wl_display *wl_display,
+        struct wlr_compositor *compositor, bool lazy);
+    void wlr_xwayland_destroy(struct wlr_xwayland *wlr_xwayland);
+    void wlr_xwayland_set_cursor(struct wlr_xwayland *wlr_xwayland,
+        uint8_t *pixels, uint32_t stride, uint32_t width, uint32_t height,
+        int32_t hotspot_x, int32_t hotspot_y);
+    void wlr_xwayland_surface_activate(struct wlr_xwayland_surface *surface,
+        bool activated);
+    void wlr_xwayland_surface_restack(struct wlr_xwayland_surface *surface,
+        struct wlr_xwayland_surface *sibling, enum xcb_stack_mode_t mode);
+    void wlr_xwayland_surface_configure(struct wlr_xwayland_surface *surface,
+        int16_t x, int16_t y, uint16_t width, uint16_t height);
+    void wlr_xwayland_surface_close(struct wlr_xwayland_surface *surface);
+    void wlr_xwayland_surface_set_minimized(struct wlr_xwayland_surface *surface,
+        bool minimized);
+    void wlr_xwayland_surface_set_maximized(struct wlr_xwayland_surface *surface,
+        bool maximized);
+    void wlr_xwayland_surface_set_fullscreen(struct wlr_xwayland_surface *surface,
+        bool fullscreen);
+    void wlr_xwayland_set_seat(struct wlr_xwayland *xwayland,
+        struct wlr_seat *seat);
+    bool wlr_surface_is_xwayland_surface(struct wlr_surface *surface);
+    struct wlr_xwayland_surface *wlr_xwayland_surface_from_wlr_surface(
+        struct wlr_surface *surface);
+    void wlr_xwayland_surface_ping(struct wlr_xwayland_surface *surface);
+    bool wlr_xwayland_or_surface_wants_focus(
+        const struct wlr_xwayland_surface *xsurface);
+    enum wlr_xwayland_icccm_input_model wlr_xwayland_icccm_input_model(
+        const struct wlr_xwayland_surface *xsurface);
+
+    typedef struct xcb_intern_atom_cookie_t {
+        ...;
+    } xcb_intern_atom_cookie_t;
+    typedef ... xcb_connection_t;
+    typedef struct xcb_intern_atom_reply_t {
+        uint8_t    response_type;
+        uint8_t    pad0;
+        uint16_t   sequence;
+        uint32_t   length;
+        xcb_atom_t atom;
+        ...;
+    } xcb_intern_atom_reply_t;
+    typedef ... xcb_generic_error_t;
+
+    int xcb_connection_has_error(xcb_connection_t *c);
+    xcb_connection_t *xcb_connect(const char *displayname, int *screenp);
+    void xcb_disconnect(xcb_connection_t *c);
+    xcb_intern_atom_cookie_t xcb_intern_atom(xcb_connection_t *conn, uint8_t only_if_exists,
+        uint16_t name_len, const char *name);
+    xcb_intern_atom_reply_t *xcb_intern_atom_reply(xcb_connection_t *conn,
+        xcb_intern_atom_cookie_t cookie, xcb_generic_error_t **e);
+    """
 
 ffi_builder = FFI()
 ffi_builder.set_source(
